# include <stdlib.h>
# include <assert.h>
# include "gcc-externs.h"
# include "sdg.h"
# include "bb.h"
# include "regset.ic"

List loopList;	/* List of all loops in the function */

static int
ComputeDominators(BasicBlock *bb, BBset *domSets)
{
	Iterator  iter;
	Edge     *edge;
	BBset     newDomSet;
	int       setChanged;

		/* newDomSet = {myself} U (Intersection(dom sets of all preds)) */
	AllocBBset(newDomSet);

		/* Loop computes Intersection(dom sets of all preds) */
	if (bb->id != 0)
		BB_SetAllBits(newDomSet);
	else {
			/* Special case that arises because of function inlining.
			 * Earlier, had the requirement that bb 0 not have any preds.
			 * But, becuase of inlining, this is no longer true */
		InitBBset(newDomSet);
		BB_SetBit(newDomSet, 0);
	}
	for (edge = (Edge *)InitIter(iter, bb->preds);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		if (!edge->isCycleEdge)
			BB_SetIntersect(newDomSet, domSets[edge->src->id]);
	}

		/* This unions {myself} with the new dom set */
	BB_SetBit(newDomSet, bb->id);

	setChanged = BB_SetsDiffer(newDomSet, domSets[bb->id]);
	if (setChanged)
		BB_CopySet(newDomSet, domSets[bb->id]);

	FreeBBset(newDomSet);

	return setChanged;
}

static void
FigureOutLoopNesting(Loop *innerLoop, Loop *newLoop, BBset *domSets)
{
/* 'innerLoop' is a loop which was identified earlier on.
 * 'newLoop' is a loop that was discovered just now.
 * Also, it is known that 'innerLoop' is nested in 'newLoop'
 * But, we need to figure if 'newLoop' encloses some other loop which
 * in turn encloses 'innerLoop' */

	Loop *oldLoop = innerLoop->enclosingLoop;

	assert(BB_BitIsSet(domSets[innerLoop->hdr->id], newLoop->hdr->id));
	if (oldLoop == NULL) {
		innerLoop->enclosingLoop = newLoop;
		return;
	}
	else if (BB_BitIsSet(domSets[newLoop->hdr->id], oldLoop->hdr->id)) {
			/* newLoop's hdr is dominated by oldLoop's hdr.
			 * So, newLoop is inside oldLoop */
		newLoop->enclosingLoop = oldLoop;
		return;
	}
	else
			/* oldLoop's hdr is dominated by newLoop's hdr.
			 * So, we know 'oldLoop' is inside 'newLoop'.
			 * But, how much inside?  So, recurse .... */
		FigureOutLoopNesting(oldLoop, newLoop, domSets);
}

static void
FindLoopBody(BasicBlock *hdr, BBset *domSets)
{
/* Find the body of the loop which has the header 'hdr'.
 * The work stack algorithm is used to traverse backward along
 * the back edges discovering the loop body. */

	Loop       *loop;
	Edge       *edge;
	List        workStack;
	BBset       bbset;
	Iterator    iter, tmpIter;
	BasicBlock *bb;

	loop = (Loop *)malloc(sizeof(Loop));
	Append(&loopList, loop);

	loop->hdr = hdr;
	Init(loop->body);
	AllocBBset(loop->backEdges);
	AllocBBset(loop->visitMask);
	InitBBset(loop->backEdges);
	InitBBset(loop->visitMask);
	loop->backEdgeProb = 0.0;
	loop->enclosingLoop = NULL;
	
	Insert(&loop->body, hdr);
	hdr->loopDepth++;
	hdr->innerLoop = loop;
	AllocBBset(bbset);
	InitBBset(bbset);
	BB_SetBit(bbset, hdr->id);
	for (edge = (Edge *)InitIter(iter, hdr->preds);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		if (edge->isBackEdge) {
			BB_SetBit(loop->backEdges, edge->src->id);

			Init(workStack);
			Insert(&workStack, edge->src);

			while ((bb = (BasicBlock *)RemoveHead(&workStack)) != NULL) {
					/* Add 'bb' to 'loop'
					 * By checking whether bit is set, make sure that 
					 * no duplicates are present in 'loop->body' */
				if (!BB_BitIsSet(bbset, bb->id)) {
					BB_SetBit(bbset, bb->id);
					Append(&loop->body, bb);
					bb->loopDepth++;
					if (bb->innerLoop != NULL) {
							/* 'bb->innerLoop' is 'oldLoop'.  If the header of
							 * 'oldLoop' dominates 'hdr', only then is 'loop'
							 * nested in 'oldLoop'.  In some rare cases, the
							 * backedges are visited in the wrong order and
							 * this results in loops being visited inside-out!
							 * In such cases, don't want to incorrectly set
							 * innerLoop.
							 */
						BasicBlock *oldHdr = bb->innerLoop->hdr;
						if (BB_BitIsSet(domSets[hdr->id], oldHdr->id)) {
							loop->enclosingLoop = bb->innerLoop;
							bb->innerLoop = loop;
						}
						else
							FigureOutLoopNesting(bb->innerLoop, loop, domSets);
					}
					else 
						bb->innerLoop = loop;
					for (edge = (Edge *)InitIter(tmpIter, bb->preds);
						 edge != NULL;
						 edge = (Edge *)GetNext(tmpIter))
					{
						if (!edge->isCycleEdge)
							Insert(&workStack, edge->src);
					}
				}
			}
		}
	}

	FreeBBset(bbset);
}

static void
MarkBackEdges()
{
	int         i; 
	BBset      *domSets;
	List 	    workList;
	BBset       bbset;
	BasicBlock *bb;

	AllocBBset(bbset);
	InitBBset(bbset);

	Init(workList);

		/* Allocate dominator sets for all basic blocks */
	domSets = (BBset *)malloc(sizeof(BBset) * n_basic_blocks);

		/* For basic block 0, set dominator to itself and nothing else */
	AllocBBset(domSets[0]);
	InitBBset(domSets[0]);
	BB_SetBit(domSets[0], 0);

	bb = GetBasicBlock(0);
	bb->loopDepth = 0;
	bb->innerLoop = NULL;

		/* Set dominators for all bbs(but bb 0) to everything.
		 * Also, put all basic blocks onto the worklist */
	for (i = 1; i < n_basic_blocks; i++) {
		AllocBBset(domSets[i]);
		bb = GetBasicBlock(i);
		bb->loopDepth = 0;
		bb->innerLoop = NULL;
		if (bb->reachable) {
				/* Initialise only if 'bb' is not a dead bb. */
			BB_SetAllBits(domSets[i]);
			Append(&workList, GetBasicBlock(i));
			BB_SetBit(bbset, i);
		}
		else
				/* For dead bbs, set dominator set to an empty set */
			InitBBset(domSets[i]);
	}

		/* Remove elements one by one and proceed */
	while ((bb = RemoveHead(&workList)) != NULL) {
		BB_UnsetBit(bbset, bb->id);
		if (ComputeDominators(bb, domSets))
			AddSuccessorsToWorkList(bb, &workList, bbset);
	}

		/* Now compute back edges and identify loop bodies
		 * Since basic blocks are visited in pgm. order, loops are
		 * discovered top-down.  The innermost loop is discovered last.
		 * This is what is needed for all basic blocks.  All bbs point
		 * to their innermost loop */
	for (i = 0; i < n_basic_blocks; i++) {
		Iterator    iter;
		BasicBlock *pred;
		Edge       *edge;
		int         isLoopHdr;

		isLoopHdr = FALSE;
		bb = GetBasicBlock(i);
		for (edge = (Edge *)InitIter(iter, bb->preds);
			 edge != NULL;
			 edge = (Edge *)GetNext(iter))
		{
			if (edge->isCycleEdge) {
				edge->isBackEdge = FALSE;
				continue;
			}

				/* 'edge' is a backedge if 'bb' dominates its pred. 'pred'
				 * i.e. the bit corr. to 'bb' is set in 'pred's domSet */
			pred = edge->src;
			edge->isBackEdge = FALSE;
			if (BB_BitIsSet(domSets[pred->id], bb->id)) {
				edge->isBackEdge = TRUE;
				isLoopHdr = TRUE;
			}
		}

		if (isLoopHdr)
			FindLoopBody(bb, domSets);
	}

		/* Free allocated memory */
	FreeBBset(bbset);
	for (i = 0; i < n_basic_blocks; i++)
		FreeBBset(domSets[i]);
	free(domSets);
}

void
FreeLoops()
{
	Loop *loop;

	for (loop = (Loop *)RemoveHead(&loopList);
		 loop != NULL;
		 loop = (Loop *)RemoveHead(&loopList))
	{
		FreeBBset(loop->backEdges);
		FreeBBset(loop->visitMask);

		while (RemoveHead(&loop->body) != NULL)
			;

		free(loop);
	}
}

int
DetectAndBreakCycle(BBset stackSet, BBset visitSet, BasicBlock *bb)
{
	Iterator    iter;
	Edge       *edge;
	BasicBlock *dst, *src;

	BB_SetBit(stackSet, bb->id);
	BB_SetBit(visitSet, bb->id);
	for (edge = (Edge *)InitIter(iter, bb->succs);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		if (!edge->isBackEdge && !edge->isCycleEdge) {
			src = edge->src;
			dst = edge->dst;
			if (BB_BitIsSet(stackSet, dst->id)) {
				fprintf(stderr, "Found cycle (%d->%d)\n", src->id, dst->id);
				edge->isCycleEdge = 1;
				return TRUE;
			}
			else if (!BB_BitIsSet(visitSet, dst->id)) {
				if (DetectAndBreakCycle(stackSet, visitSet, dst) == TRUE)
					return TRUE;
			}
		}
	}
	BB_UnsetBit(stackSet, bb->id);

	return FALSE;
}

void
BreakCyclesIfGraphIsNonreducible()
{
	BBset stackSet, visitSet;
	int   found, cyclePresent;

	AllocBBset(visitSet);
	AllocBBset(stackSet);

	cyclePresent = FALSE;
	do {
		InitBBset(visitSet);
		InitBBset(stackSet);
		found = DetectAndBreakCycle(stackSet, visitSet, GetBasicBlock(0));
		if (found == TRUE)
			cyclePresent = TRUE;
	} while(found == TRUE);

	FreeBBset(stackSet);
	FreeBBset(visitSet);

	if (cyclePresent == TRUE) {
		FreeLoops();
		MarkBackEdges();
	}
}

static void
dfs(BasicBlock *bb)
{
	Edge     *edge;
	Iterator  iter;

	bb->reachable = 1;
	for (edge = (Edge *)InitIter(iter, bb->succs);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		if (edge->dst->reachable == 0)
			dfs(edge->dst);
	}
}

static void
DetectReachableBBs()
{
	int i;

	dfs(GetBasicBlock(0));
	for (i = 0; i < n_basic_blocks; i++)
		if (GetBasicBlock(i)->reachable == 0)
			fprintf(stderr, "Basic block %d is dead\n", i);
}

# ifdef __GNUC__
inline
# endif
static float
ComputeExitProbAlongExit(BasicBlock *exit)
{
/*
 * Before exiting, accumulate probabilities along
 * all back edges at the header.  This is used to
 * compute the probability of exiting out of this
 * loop along this edge using the formula below
 * 
 * p1 = Prob. of entering the loop
 *    = Prob. of loop hdr 
 *    = loop->hdr->prob
 * p2 = Prob. of executing the loop a second time
 *    = Sum of probs. along back edges
 *    = loop->backEdgeProb
 * p3 = Prob. of exiting along this edge the
 *       first time around
 *    = src->succProb
 * Then prob. of exiting this loop along this
 * edge is p = p3 * p1 / (p1 - p2)
 */

	Loop *loop = exit->innerLoop;

	return (exit->succProb / (1.0 - loop->backEdgeProb/loop->hdr->prob));
}

static BasicBlock *
CreateDummyExitOutOfLoop(BasicBlock *exit)
{
/* 'exit' is a bb along which there is an edge out of the loop
 * But, this edge jumps out of more than 1 loop.  So, create a
 * dummy exit which takes us out of exactly 1 loop */

	BasicBlock *dummyExit;

	fprintf(stderr, "Creating dummy exit for loop with hdr %d at exit %d\n",
					 exit->innerLoop->hdr->id, exit->id);
	dummyExit = (BasicBlock *)malloc(sizeof(BasicBlock));
	dummyExit->loopDepth = exit->loopDepth - 1;
	dummyExit->innerLoop = exit->innerLoop->enclosingLoop;
	dummyExit->prob = ComputeExitProbAlongExit(exit);
	dummyExit->succProb = dummyExit->prob;
	dummyExit->dummyExit = NULL;
	dummyExit->id = -1;

	return dummyExit;
}

void
ComputeBBexecutionProbabilities()
{
/* Do a breadth first traversal of the basic blocks and assign
 * probabilities of this basic block executing.
 * For now, this is done by assuming all branches are equally likely
 * to be taken.  This can also be done based on profiling.. */

	List        bbList;
	Iterator    iter;
	Edge       *edge;
	BasicBlock *bb, *src, *dst;
	int         numSuccs;
	BBset       bbset;

/*
		 * ASSUMPTION : Basic block '0', the first basic block in this proc
		 * does not have any predecessors.  This assumption is required 
		 * for correct computation of dominator information.
	assert(IsEmpty(bb->preds));
************
* SSS : As it turns out, because of inlining, bb 0 can have some predecessors
* One solution to this problem is to insert a dummy BB in the graph
* right in the beginning, but that would require changing the interation
* of this code and gcc.  gcc does not insert such a dummy bb and I use the
* information computed by gcc to do my stuff.
* Instead, am adding a check for bb 0 in the dominator algorithm and 
* special-casing it for bb 0.
************
*/
	DetectReachableBBs();
	MarkBackEdges();
	BreakCyclesIfGraphIsNonreducible();

	Init(bbList);
	bb = GetBasicBlock(0);
	bb->prob = 1.0;

	AllocBBset(bbset);
	InitBBset(bbset);

		/* Initialise the list of bbs to visit */
	numSuccs = 0;
	for (edge = (Edge *)InitIter(iter, bb->succs);
		 edge != NULL;
		 edge = (Edge *)GetNext(iter))
	{
		if (!edge->isCycleEdge) {
			numSuccs++;
			Append(&bbList, edge->dst);
			BB_SetBit(bbset, edge->dst->id);
		}
	}

	if (numSuccs > 0)
		bb->succProb = (bb->prob/numSuccs);

		/* Sit in a loop and process all basic blocks one-by-one
		 * till we are done.  Essentially, we are doing some sort
		 * of a breadth-first search. */
	for (bb = (BasicBlock *)RemoveHead(&bbList); 
	     bb != NULL; 
		 bb = (BasicBlock *)RemoveHead(&bbList))
	{
		float prob = 0.0;

		BB_UnsetBit(bbset, bb->id);
		for (edge = (Edge *)InitIter(iter, bb->preds);
			 edge != NULL;
			 edge = (Edge *)GetNext(iter))
		{
				/* Back edges are ignored for probability computations */
			if (edge->isBackEdge || edge->isCycleEdge)
				continue;
			
			src = edge->src;
			dst = edge->dst;

					/* Skip 'bb' and wait for probability of 'src'
					 * to be computed */
			if (src->prob == -1.0)
				goto skipBB;

					/* Check if 'edge' is NOT a loop exit edge */
			if (src->loopDepth <= dst->loopDepth)
				prob += src->succProb;
			else {
				while (1) {
						/* 'src->dst' takes us out of a loop(possibly > 1)
						 * So, compute the prob. of exiting along this edge.
						 * Before anything can be done, probs along all back
						 * edges of the loop should have been computed.  Check
						 * for that */
					Loop *loop = src->innerLoop;
					if (BB_SetsDiffer(loop->visitMask, loop->backEdges)) {
						Append(&bbList, bb);
						BB_SetBit(bbset, bb->id);
						goto skipBB;
					}

						/* Check if 'src->dst' jumps out of exactly 1 loop */
					if (src->loopDepth == (dst->loopDepth + 1))
						break;

						/* Since 'src->dst' takes us out of more than 1 loop,
						 * create a dummy exit to jump out of exactly 1 loop */
					if (src->dummyExit == NULL)
						src->dummyExit = CreateDummyExitOutOfLoop(src);

					src = src->dummyExit;
				}

				prob += ComputeExitProbAlongExit(src);
			}
		}

		if (prob > 1.0) {
			fprintf(stderr, "**** BASIC BLOCK %d : %1.3f ****\n", bb->id, prob);
			bb->prob = 1.0;
		}
		else
			bb->prob = prob;

			/* Count the # of successors of this bb and
			 * compute the successor prob */
		numSuccs = 0;
		for (edge = (Edge *)InitIter(iter, bb->succs);
			 edge != NULL;
			 edge = (Edge *)GetNext(iter))
		{
			if (!edge->isCycleEdge)
				numSuccs++;
		}

		if (numSuccs > 0)
			bb->succProb = bb->prob/numSuccs;

		for (edge = (Edge *)InitIter(iter, bb->succs);
			 edge != NULL;
			 edge = (Edge *)GetNext(iter))
		{
			if (edge->isBackEdge) {
					/* Set bit in the visit mask of the containing loop.
					 * Use 'edge->dst' because there can be multiple back 
					 * edges originating from 'edge->src' to different loops
					 * Also, accumulate to the total loop backedge prob */
				Loop *loop = edge->dst->innerLoop;
				BB_SetBit(loop->visitMask, edge->src->id);
				loop->backEdgeProb += bb->succProb;
			}
			else if (!edge->isCycleEdge) {
				dst = edge->dst;
				if (dst->prob == -1.0 && !BB_BitIsSet(bbset, dst->id)) {
					Append(&bbList, dst);
					BB_SetBit(bbset, dst->id);
				}
			}
		}

	skipBB :
	}

	FreeBBset(bbset);
	FreeLoops();
}
