# include <assert.h>
# include "gcc-externs.h"
# include "sdg.h"
# include "ud.h"
# include "bb.h"
# include "regset.ic"

# define DEFN_CHUNK_SIZE 		  50
# define DEFN_PTRS_ARRAY_SIZE     DEFN_CHUNK_SIZE
# define DEFN_PTR_PTRS_ARRAY_SIZE DEFN_CHUNK_SIZE
# define FIRST_THRESHOLD          DEFN_CHUNK_SIZE
# define SECOND_THRESHOLD                \
	(FIRST_THRESHOLD + (DEFN_PTRS_ARRAY_SIZE * DEFN_CHUNK_SIZE))
# define FINAL_THRESHOLD          		 \
	(SECOND_THRESHOLD +                  \
		(DEFN_PTR_PTRS_ARRAY_SIZE * DEFN_PTRS_ARRAY_SIZE * DEFN_CHUNK_SIZE))
# define ALLOC_DEFN_CHUNK(chunkPtr)      \
	chunkPtr = (Defn *)calloc(DEFN_CHUNK_SIZE, sizeof(Defn))
# define ALLOC_ARRAY_OF_DEFN_PTRS(array) \
	array = (Defn **)calloc(DEFN_PTRS_ARRAY_SIZE, sizeof(Defn *))

		/* Count of the total number of definitions in the function */
int numDefns;

static int 	   startDefnId;
static Defn   *defns;
static Defn  **defnPtrs;
static Defn ***defnPtrPtrs;

static rtx     currentInsn;

	/* For an instruction in basic block 'bb'
	 * nodeWeight = prob(bb) * 5^(loop-nesting-level of bb) */
static float   nodeWeight;
static int     inProfile = 0;

void
InitDefns()
{
	numDefns = 0;
	ALLOC_DEFN_CHUNK(defns);
	ALLOC_ARRAY_OF_DEFN_PTRS(defnPtrs);
	defnPtrPtrs = (Defn ***)calloc(DEFN_PTR_PTRS_ARRAY_SIZE, sizeof(Defn **));
}

Defn *
GetDefn(int defnId)
{
/* The definition structures are organised in a manner
 * analogous to the data block organisation of a file in
 * the UNIX file system.
 */

	int tmp, index, newIndex;

	if (defnId < FIRST_THRESHOLD) 				/* FIRST LEVEL */
		return &defns[defnId];
	else if (defnId < SECOND_THRESHOLD) { 	/* SECOND LEVEL */
		tmp = defnId - FIRST_THRESHOLD;
		index = tmp/DEFN_CHUNK_SIZE;
		return &defnPtrs[index][tmp - index*DEFN_CHUNK_SIZE];
	}
	else {										/* THIRD LEVEL */
		tmp = defnId - SECOND_THRESHOLD;
		index = tmp/(DEFN_CHUNK_SIZE*DEFN_PTRS_ARRAY_SIZE);
		tmp = tmp - index*DEFN_CHUNK_SIZE*DEFN_PTRS_ARRAY_SIZE;
		newIndex = tmp/DEFN_CHUNK_SIZE;
		tmp = tmp - newIndex*DEFN_CHUNK_SIZE;
		return &defnPtrPtrs[index][newIndex][tmp];
	}
}

void
AddNewDefn(int bbId, int destReg, rtx definingInstr)
{
/* The definition structures are organised in a manner
 * analogous to the data block organisation of a file in
 * the UNIX file system.
 */

	Defn *defn;
	int   tmp, index, newIndex;
	int   defnId = numDefns;

	if (destReg == STACK_POINTER_REGNUM || destReg == FRAME_POINTER_REGNUM)
		return;

	numDefns++;
	if (defnId < FIRST_THRESHOLD) 			/* FIRST LEVEL */
		defn = &defns[defnId];
	else if (defnId < SECOND_THRESHOLD) { 	/* SECOND LEVEL */
		tmp = defnId - FIRST_THRESHOLD;
		index = tmp/DEFN_CHUNK_SIZE;
		if (defnPtrs[index] == NULL)
			ALLOC_DEFN_CHUNK(defnPtrs[index]);
		defn = &defnPtrs[index][tmp - index*DEFN_CHUNK_SIZE];
	}
	else if (defnId < FINAL_THRESHOLD) {	/* THIRD LEVEL */
		tmp = defnId - SECOND_THRESHOLD;
		index = tmp/(DEFN_CHUNK_SIZE*DEFN_PTRS_ARRAY_SIZE);
		tmp = tmp - index*DEFN_CHUNK_SIZE*DEFN_PTRS_ARRAY_SIZE;
		if (defnPtrPtrs[index] == NULL) {
				/* Firstly, allocate the array of pointers to Defns 
				 * at 'defnPtrPtrs[index]'. 
				 * Next, allocate an array of Defns */
			ALLOC_ARRAY_OF_DEFN_PTRS(defnPtrPtrs[index]);
			ALLOC_DEFN_CHUNK(defnPtrPtrs[index][0]);
			defn = &defnPtrPtrs[index][0][0];
		}
		else {
			newIndex = tmp/DEFN_CHUNK_SIZE;
			if (defnPtrPtrs[index][newIndex] == NULL)
				ALLOC_DEFN_CHUNK(defnPtrPtrs[index][newIndex]);
			defn = &defnPtrPtrs[index][newIndex][tmp-newIndex*DEFN_CHUNK_SIZE];
		}
	}
	else {
		fprintf(stderr, "Far too many defns.(> %d) to handle\n", defnId);
		abort();
	}
	defn->id = defnId;
	defn->node = currSdgNode;
	defn->insn = definingInstr;
	defn->regNo = destReg;
	defn->bbId = bbId;

			/* Set bit in 'regsDefinedInBB[bbId]' corr. to 'destReg' */
	SetBit(regsDefinedInBB[bbId], destReg);
			/* This could be the first definition in the current bb
			 * if 'startDefnId' is not yet set */
	if (startDefnId == -1) 
		startDefnId = defnId;
}

void
SplitNodeIfNeeded(rtx body)
{
		/* Is a load operation */
	RTX_CODE code = GET_CODE(SET_SRC(body));

	if (   (code == MEM) || (code == SIGN_EXTEND)
		|| (code == ZERO_EXTEND) || (code == FLOAT_EXTEND)
		|| (code == FLOAT_TRUNCATE) || (code == TRUNCATE))
	{
			/* Allocate new node for the defn and add this new 
			 * node into the graph.  Now, the earlier node
			 * represents the address of the load */
		SdgNode *defnNode = AllocNewSDGnode(currSdgNode->instr);
		defnNode->next = currSdgNode->next;
		currSdgNode->next = defnNode;
		if (sdg.tail == currSdgNode)
			sdg.tail = defnNode;

		defnNode->inCompoundRtx = currSdgNode->inCompoundRtx;
		defnNode->compoundRtx = currSdgNode->compoundRtx;
		defnNode->nodeWeight = nodeWeight;
		defnNode->inProfile = inProfile;

			/* Mark 'currSdgNode as being an address node */
		currSdgNode->isAddr = 1;
		currSdgNode = defnNode;
	}
	else if (   code == FIX || code == FLOAT 
		     || code == UNSIGNED_FIX || code == UNSIGNED_FLOAT)
	{
			/* Alloc new node for the defn and add it into the graph */
		SdgNode *defnNode = AllocNewSDGnode(currSdgNode->instr);
		defnNode->next = currSdgNode->next;
		currSdgNode->next = defnNode;
		if (sdg.tail == currSdgNode)
			sdg.tail = defnNode;

		defnNode->inCompoundRtx = currSdgNode->inCompoundRtx;
		defnNode->compoundRtx = currSdgNode->compoundRtx;
		defnNode->nodeWeight = nodeWeight;
		defnNode->inProfile = inProfile;

			/* Always mark the 'fp' part of the conversion as 
			 * being the 'conversion' operation so that the 
			 * marking procedure('MarkCompInstructions') knows to skip
			 * that node.  Whether this instr. will be in 'LdSt' or in 'Comp'
			 * is really dependent on the integer part of the conversion */
		if (code == FLOAT || code == UNSIGNED_FLOAT) {
			defnNode->isAConvNode = 1;
			currSdgNode->isAConvNode = 0;
		}

		currSdgNode = defnNode;
	}
}

void
AddDefnsFrom_PARALLEL_rtx(int bbId, rtx insn)
{
/* PARALLEL holds a vector of expressions each
 * of which is a side-effect expression */

	int i;
	rtx expr, dest;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		expr = XVECEXP(insn, 0, i);
			/* Allocate a new sdg node for 'expr' */
		currSdgNode = InsertInsnIntoSDG(expr);
		currSdgNode->nodeWeight = nodeWeight;
		currSdgNode->inProfile = inProfile;
		currSdgNode->inCompoundRtx = 1;
		currSdgNode->compoundRtx = currentInsn;
		if (currBB->startNode == NULL)
			currBB->startNode = currSdgNode;

		if (GET_CODE(expr) == SET) {
			if (GET_CODE(SET_SRC(expr)) == CALL)
				currSdgNode->isACall = 1;
			else 
				SplitNodeIfNeeded(expr);

			dest = SET_DEST(expr);
			switch (GET_CODE(dest)) {
				case REG :
					AddNewDefn(bbId, REGNO(dest), expr);
					break;

				case SUBREG :
					AddNewDefn(bbId, REGNO(SUBREG_REG(dest)), expr);
					break;
				
				case STRICT_LOW_PART :
					AddNewDefn(bbId, REGNO(SUBREG_REG(XEXP(dest, 0))), expr);
					break;
			
				default :
					break;
			}
		}
		else if (GET_CODE(expr) == CALL)
			currSdgNode->isACall = 1;
	}
}

void
AddDefnsFrom_SEQUENCE_rtx(int bbId, rtx seq)
{
/* SEQUENCE holds a vector of instructions */
	void AddDefnsFromInstr(int bbId, rtx insn);

	int 	 i;
	rtx 	 insn;
	RTX_CODE code;

	for (i = 0; i < XVECLEN(seq, 0); i++) {
		insn = XVECEXP(seq, 0, i);
		code = GET_CODE(insn);

			/* Allocate a new sdg node for 'insn' */
		if (code == INSN || code == CALL_INSN || code == JUMP_INSN) {
			currSdgNode = InsertInsnIntoSDG(insn);
			currSdgNode->nodeWeight = nodeWeight;
			currSdgNode->inProfile = inProfile;
			currSdgNode->inCompoundRtx = 1;
			currSdgNode->compoundRtx = currentInsn;
			if (currBB->startNode == NULL)
				currBB->startNode = currSdgNode;
		}

		if (code == INSN || code == CALL_INSN)
			AddDefnsFromInstr(bbId, insn);
	}
}

void
AddDefnsFrom_UNSPEC_rtx(int bbId, rtx insn)
{
/* UNSPEC/UNSPEC_VOLATILE have operands which're used in some
 * m/c specific operation(s).  These are considered to define/use
 * all their operands */

	int i;
	rtx elt;

	for (i = 0; i < XVECLEN(insn, 0); i++) {
		elt = XVECEXP(insn, 0, i);
		if (REG_P(elt))
			AddNewDefn(bbId, REGNO(elt), insn);
	}
}

void
AddDefnsFromInstr(int bbId, rtx insn)
{
	rtx body = PATTERN(insn);
	rtx dest;

	switch (GET_CODE(body)) {
	case RETURN :
	case USE :
	case ADDR_VEC :
	case ADDR_DIFF_VEC :
			/* In all these cases, there is no new defn. */
			/* Just return without doing nothing */
		return;

	case CALL :
			/* Returns no value => no new defn. */
		currSdgNode->isACall = 1;
		return;

	case PARALLEL :
		AddDefnsFrom_PARALLEL_rtx(bbId, body);
		break;

	case SEQUENCE :
		AddDefnsFrom_SEQUENCE_rtx(bbId, body);
		break;

	case CLOBBER :
		dest = XEXP(body, 0);
		if (REG_P(dest))
			AddNewDefn(bbId, REGNO(dest), body);
		break;

	case SET :
		SplitNodeIfNeeded(body);
		dest = SET_DEST(body);
		switch (GET_CODE(dest)) {
			case REG :
				AddNewDefn(bbId, REGNO(dest), body);
				break;

			case SUBREG :
				AddNewDefn(bbId, REGNO(SUBREG_REG(dest)), body);
				break;
			
			case STRICT_LOW_PART :
				AddNewDefn(bbId, REGNO(SUBREG_REG(XEXP(dest, 0))), body);
				break;
		
			default :
				break;
		}
		break;

	case UNSPEC :
	case UNSPEC_VOLATILE :
			/* These are considered to set/use all its operands */
		AddDefnsFrom_UNSPEC_rtx(bbId, body);
		break;

	case TRAP_IF :
		break;

	case ASM_INPUT :
/*
	case ASM_OUTPUT :
*/
		fprintf(stderr, "Inline ASM not yet handled...\n");
		abort();

	default :
		fprintf(stderr, "Unhandled case in \'AddDefnsFromInstr\'\n");
		abort();
	}
}

# define MAX_LOOP_WEIGHT 100
# define min(a,b) (((a) < (b)) ? (a) : (b))
# define max(a,b) (((a) > (b)) ? (a) : (b))

int pow(int a, int b)
{
	int prod = 1;
	for ( ; b > 0; b--)
		prod *= a;

	return prod;
}

void
IdentifyDefnsInBB(int bbId)
{
	short       tmp;
	rtx         first;
	rtx         last;
	rtx         insn;
	rtx         next;
	RTX_CODE    code;

	first = basic_block_head[bbId]; 
	last = basic_block_end[bbId];
	currBB = GetBasicBlock(bbId); 
	startDefnId = -1;

		/* Initialise the regset 'regsDefinedInBB' */
	InitRegset(regsDefinedInBB[bbId]);

# ifndef PLAIN_SLICING
	{
		extern unsigned long *profile_BBcount;

		nodeWeight = currBB->prob * min(MAX_LOOP_WEIGHT, 
										(int)pow(5, currBB->loopDepth)); 
		nodeWeight = max(nodeWeight, profile_BBcount[bbId]);
		inProfile  = (profile_BBcount[bbId] != 0);
	}
# endif
# ifdef COPYOPT_SLICING
	if (bbId == 0) {
		int i;
		for (i = 0; i < GP_ARG_LAST - GP_ARG_FIRST + 1; i++) {
			dummyArgNodes[i]->nodeWeight = nodeWeight;
			dummyArgNodes[i]->inProfile = 0;
		}
	}
# endif
	for (insn = first; ; insn = next) {
		next = NEXT_INSN(insn);
		code = GET_CODE(insn);

		currentInsn = insn;

		switch (code) {
				/* Notes, barriers and labels are useless... */
				/* They cant result in any new defns. */
			case NOTE :
			case BARRIER :
			case CODE_LABEL :
				break;

				/* For JUMP_INSN, INSN, CALL_INSN, identify possible defns. */
			case JUMP_INSN :
			case CALL_INSN :
			case INSN :
				if (   (GET_CODE(PATTERN(insn)) != PARALLEL)
					&& (GET_CODE(PATTERN(insn)) != SEQUENCE))
				{
					currSdgNode = InsertInsnIntoSDG(insn);
					currSdgNode->nodeWeight = nodeWeight;
					currSdgNode->inProfile = inProfile;
					if (currBB->startNode == NULL)
						currBB->startNode = currSdgNode;
				}

				AddDefnsFromInstr(bbId, insn);
				break;

			default :
				fprintf(stderr, "Unhandled case in \'IdentifyDefnsInBB\'\n");
				abort();
		}

		if (insn == last)
			break;
	}

	currBB->startDefnId = startDefnId; 
}

void
PrintDefns()
{
	int   i;
	Defn *defn;

	extern void print_single_rtl(FILE *, rtx);

	for (i = 0; i < numDefns; i++) {
		defn = GetDefn(i);
		fprintf(stderr, "## Defn %d ##\n", i);
		print_single_rtl(stderr, defn->insn);
		fputc('\n', stderr);
	}
}

void
FreeDefns()
{
	int i, j;

		/* FREE FIRST LEVEL */
	free(defns);
	defns = NULL;

		/* FREE SECOND LEVEL */
	for (i = 0; i < DEFN_PTRS_ARRAY_SIZE; i++)
		if (defnPtrs[i] == NULL)
			break;
		else
			free(defnPtrs[i]);

	free(defnPtrs);
	defnPtrs = NULL;

		/* FREE THIRD LEVEL */
	for (i = 0; i < DEFN_PTR_PTRS_ARRAY_SIZE; i++)
		if (defnPtrPtrs[i]) {
			for (j = 0; j < DEFN_PTRS_ARRAY_SIZE; j++)
				if (defnPtrPtrs[i][j] == NULL)
					break;
				else
					free(defnPtrPtrs[i][j]);

			free(defnPtrPtrs[i]);
		}

	free(defnPtrPtrs);
	defnPtrPtrs = NULL;
}
